import { Component, ViewChild, AfterContentInit, ChangeDetectorRef } from "@angular/core";

import { ChartComponent } from "angular2-chartjs";

import { Observer, ParamFamily } from "jalhyd";

import { GraphTypeSelectComponent } from "./graph-type.component";
import { ApplicationSetupService } from "../../services/app-setup/app-setup.service";
import { I18nService } from "../../services/internationalisation/internationalisation.service";
import { PlottableData } from "../../results/plottable-data";
import { GraphType } from "../../results/graph-type";
import { ResultsComponent } from "../fixedvar-results/results.component";
import { IYSeries } from "../../results/y-series";
import { VarResults } from "../../results/var-results";

@Component({
    selector: "results-graph",
    templateUrl: "./results-graph.component.html",
    styleUrls: [
        "./results-graph.component.scss"
    ]
})
export class ResultsGraphComponent extends ResultsComponent implements AfterContentInit, Observer {

    @ViewChild(ChartComponent, { static: false })
    private chartComponent;

    private _results: PlottableData;

    /** used to briefly destroy/rebuild the chart component, to refresh axis labels (@see bug #137) */
    public displayChart = true;

    private _zoomWasChanged = false;

    @ViewChild(GraphTypeSelectComponent, { static: true }) // true, or else adding observer will fail in ngAfterContentInit()
    private _graphTypeComponent: GraphTypeSelectComponent;

    /*
     * config du graphe
     */
    public graph_type: string;
    public graph_data: any = {};
    public graph_options: any = {
        responsive: true,
        maintainAspectRatio: true,
        animation: {
            duration: 0
        },
        legend: {
            display: false
        },
        title: {
            display: true,
            text: ""
        },
        elements: {
            line: {
                tension: 0
            }
        }
    };

    public constructor(
        private appSetup: ApplicationSetupService,
        private intlService: I18nService,
        private cd: ChangeDetectorRef
    ) {
        super();
        // enable zoom and pan (using "chartjs-plugin-zoom" package)
        const that = this;
        this.graph_options["plugins"] = {
            zoom: {
                pan: {
                    enabled: false, // conflicts with drag zoom
                    mode: "xy",
                },
                zoom: {
                    enabled: true,
                    drag: { // conflicts with pan; set to false to enable mouse wheel zoom,
                        borderColor: "rgba(225,225,225,0.3)",
                        borderWidth: 1,
                        backgroundColor: "rgba(0,0,0,0.25)"
                    },
                    mode: "xy",
                    // percentage of zoom on a wheel event
                    // speed: 0.1,
                    onZoomComplete: function(t: any) { return function() { t.zoomComplete(); }; }(that)
                }
            }
        };
    }

    public set results(r: PlottableData) {
        this.forceRebuild();
        this._results = r;
        if (this._results && this._graphTypeComponent) {
            this._graphTypeComponent.selectedValue = r.graphType;
        }
    }

    public get availableXAxis() {
        if (this._results) {
            return this._results.getAvailableXAxis();
        }
    }

    public get availableYAxis() {
        if (this._results) {
            if (this._results.graphType !== GraphType.Scatter) {
                // do not use real Y axis (that include families), if chart cannot display multiple series
                return this._results.getAvailableXAxis();
            } else {
                return this._results.getAvailableYAxis();
            }
        }
    }

    public get chartX() {
        if (this._results) {
            return this._results.chartX;
        }
    }

    public set chartX(X) {
        if (X !== this.chartX) {
            this._results.chartX = X;
            this.forceRebuild();
            // refresh chart
            this.updateView();
        }
    }

    public get chartY() {
        if (this._results) {
            return this._results.chartY;
        }
    }

    public set chartY(Y) {
        if (Y !== this.chartY) {
            this._results.chartY = Y;
            this.forceRebuild();
            // refresh chart
            this.updateView();
        }
    }

    /**
     * Returns a human readable description of any param / result symbol
     */
    public getChartAxisLabel(symbol: string): string {
        return this._results.getChartAxisLabel(symbol);
    }

    public get uitextSelectX() {
        return this.intlService.localizeText("INFO_PARAMFIELD_GRAPH_SELECT_X_AXIS");
    }

    public get uitextSelectY() {
        return this.intlService.localizeText("INFO_PARAMFIELD_GRAPH_SELECT_Y_AXIS");
    }

    private zoomComplete() {
        this._zoomWasChanged = true;
        this.cd.detectChanges();
    }

    public get zoomWasChanged(): boolean {
        return this._zoomWasChanged;
    }

    public updateView() {
        // (re)generate chart
        switch (this._graphTypeComponent.selectedValue) {
            case GraphType.Histogram:
                this.graph_type = "bar";
                this.generateBarGraph();
                break;

            default:
                this.graph_type = "scatter";
                this.generateScatterGraph();
                break;
        }
    }

    public ngAfterContentInit() {
        this._graphTypeComponent.addObserver(this);
    }

    /**
     * Calls getChartAxisLabel() and removes the symbol prefix
     * (cannot rebuild a clean label here)
     */
    private axisLabelWithoutSymbol(symbol: string) {
        // detect children results
        const match = /^([0-9]+)_(.+)$/.exec(symbol);
        // get label with symbol
        let l = this._results.getChartAxisLabel(symbol);
        // remove symbol
        if (match === null) { // child prefix also uses ":"
            const i = l.indexOf(": ");
            if (i !== -1) {
                l = l.substring(i + 2);
            }
        }
        return l;
    }

    /** forces Angular to rebuild the chart @see bug #137 */
    private forceRebuild() {
        this.displayChart = false;
        const that = this;
        setTimeout(() => { // trick
            that.displayChart = true;
        }, 10);
    }

    /**
     * génère les données d'un graphe de type "bar"
     */
    private generateBarGraph() {
        const labs = [];
        const dat = [];
        const xSeries = this._results.getValuesSeries(this.chartX);
        const ySeries = this._results.getValuesSeries(this.chartY);

        // both series are supposed to be the same length
        for (let i = 0; i < xSeries.length; i++) {
            labs.push((xSeries[i] !== undefined) ? xSeries[i] : "");
            dat.push((ySeries[i] !== undefined) ? ySeries[i] : "");
        }

        const nDigits = this.appSetup.displayDigits;
        this.graph_options["scales"] = {
            xAxes: [{
                gridLines: {
                    offsetGridLines: true
                },
                ticks: {
                    precision: nDigits,
                    callback: function(value, index, values) {
                        return Number(value).toFixed(nDigits);
                    }
                },
                scaleLabel: {
                    display: true,
                    labelString: this.axisLabelWithoutSymbol(this.chartX)
                }
            }],
            yAxes: [{
                scaleLabel: {
                    display: true,
                    labelString: this.axisLabelWithoutSymbol(this.chartY)
                }
            }]
        };

        const that = this;
        this.graph_options["tooltips"] = {
            displayColors: false,
            callbacks: {
                title: (tooltipItems, data) => {
                    return this.chartY + " = " + Number(tooltipItems[0].yLabel).toFixed(nDigits);
                },
                label: (tooltipItem, data) => {
                    const lines: string[] = [];
                    const nbLines = that._results.getVariatingParametersSymbols().length;
                    for (const v of that._results.getVariatingParametersSymbols()) {
                        const series = that._results.getValuesSeries[0](v);
                        const line = v + " = " + series[tooltipItem.index].toFixed(nDigits);
                        if (v === this.chartX) {
                            if (nbLines > 1) {
                                lines.unshift("");
                            }
                            lines.unshift(line);
                        } else {
                            lines.push(line);
                        }
                    }
                    return lines;
                }
            }
        };

        this.graph_data = {
            labels: labs,
            datasets: [{
                label: "",
                data: dat
            }]
        };
    }

    /**
     * génère les données d'un graphe de type "scatter"
     */
    private generateScatterGraph() {
        this.graph_data = {
            datasets: []
        };
        const nDigits = this.appSetup.displayDigits;
        const ySeries = this.getYSeries(this.chartY);

        // are we dealing with multiple Y series ?
        const isMultiple = (ySeries.length > 1);

        // all series are supposed to be the same length
        for (const ys of ySeries) {
            this.graph_data.datasets.push({
                label: ys.label,
                data: ys.data,
                borderColor: ys.color,
                backgroundColor: "rgba(0,0,0,0)",  // fill color under the line : transparent
                showLine: "true"
            });
        }

        this.graph_options["scales"] = {
            xAxes: [{
                type: "linear",
                position: "bottom",
                ticks: {
                    precision: nDigits
                },
                scaleLabel: {
                    display: true,
                    labelString: this.axisLabelWithoutSymbol(this.chartX)
                }
            }],
            yAxes: [{
                type: "linear",
                position: "left",
                ticks: {
                    precision: nDigits
                },
                scaleLabel: {
                    display: true,
                    labelString: this.axisLabelWithoutSymbol(this.chartY)
                }
            }]
        };

        if (isMultiple) {
            // add legend for multiple series
            this.graph_options.legend = {
                display: true,
                position: "bottom",
                reverse: false
            };
            // remove tooltips @TODO make them work for multiple series
            delete this.graph_options.tooltips;
        } else {
            // enhanced tooltips for single series
            const that = this;
            this.graph_options.tooltips = {
                displayColors: false,
                callbacks: {
                    title: (tooltipItems, data) => {
                        return this.chartY + " = " + Number(tooltipItems[0].yLabel).toFixed(nDigits);
                    },
                    label: (tooltipItem, data) => {
                        let lines: string[] = [];
                        // 1. X if different from Y
                        if (this.chartX !== this.chartY) {
                            const xseries = that._results.getValuesSeries(this.chartX);
                            const xline = this.chartX + " = " + xseries[tooltipItem.index].toFixed(nDigits);
                            lines.push(xline);
                        }
                        // 2. variated parameters other than X or Y
                        const varLines: string[] = [];
                        for (const v of that._results.getVariatingParametersSymbols()) {
                            if (v !== this.chartX && v !== this.chartY) {
                                const series = that._results.getValuesSeries(v);
                                const line = v + " = " + series[tooltipItem.index].toFixed(nDigits);
                                varLines.push(line);
                            }
                        }
                        // blank line ?
                        if (varLines.length > 0) {
                            lines.push("");
                            lines = lines.concat(varLines);
                        }
                        return lines;
                    }
                }
            };
            // remove legend
            this.graph_options.legend = {
                display: false
            };
        }
    }

    /**
     * Returns a list of plottable data series for the given symbol; unless symbol
     * is a ParamFamily, the returned list will have only one value
     */
    private getYSeries(symbol: string): IYSeries[] {
        const ret: IYSeries[] = [];
        const palette = ResultsComponent.distinctColors;
        const xSeries = this._results.getValuesSeries(this.chartX);
        const nDigits = this.appSetup.displayDigits;
        let symbols: string[];

        // whole family of variables => multiple series (should only happen with VarResults)
        if (ParamFamily[symbol] !== undefined && this._results instanceof VarResults) {
            symbols = this._results.extractFamilies()[symbol];
        } else {
            symbols = [ symbol ];
        }
        // loop over found symbol(s)
        let i = 0;
        for (const s of symbols) {
            const dat = [];
            const ySeries = this._results.getValuesSeries(s);
            // build fixed precision x/y coordinate pairs
            for (let j = 0; j < xSeries.length; j++) {
                dat.push({
                    x: (xSeries[j] !== undefined) ? xSeries[j].toFixed(nDigits) : "",
                    y: (ySeries[j] !== undefined) ? ySeries[j].toFixed(nDigits) : ""
                });
            }
            // add series
            ret.push({
                data: dat,
                label: this.axisLabelWithoutSymbol(s),
                color: palette[ i % palette.length ]
            });
            i++;
        }

        return ret;
    }

    public exportAsImage(element: HTMLDivElement) {
        const canvas: HTMLCanvasElement = element.querySelector("canvas");
        canvas.toBlob((blob) => {
            saveAs(blob, "chart.png");
        }); // defaults to image/png
    }

    public resetZoom() {
        this.chartComponent.chart.resetZoom();
        this._zoomWasChanged = false;
    }

    public get uitextResetZoomTitle() {
        return this.intlService.localizeText("INFO_GRAPH_BUTTON_TITLE_RESET_ZOOM");
    }

    public get uitextExportImageTitle() {
        return this.intlService.localizeText("INFO_GRAPH_BUTTON_TITLE_EXPORT_IMAGE");
    }

    public get uitextEnterFSTitle() {
        return this.intlService.localizeText("INFO_GRAPH_BUTTON_TITLE_ENTER_FS");
    }

    public get uitextExitFSTitle() {
        return this.intlService.localizeText("INFO_GRAPH_BUTTON_TITLE_EXIT_FS");
    }

    // interface Observer

    update(sender: any, data: any) {
        if (sender instanceof GraphTypeSelectComponent) {
            if (data.action === "graphTypeChanged") {
                this._results.graphType = data.value;
            }
        }
        this.updateView();
    }
}
