import { CalculatorType, ComputeNodeType, ParamDefinition, LoiDebit, Props, Observer, Nub, MethodeResolution, StructureType } from "jalhyd";

import { FormulaireElement } from "./formulaire-element";
import { Field } from "./field";
import { SelectField } from "./select-field";
import { NgParameter, ParamRadioConfig } from "./ngparam";
import { FormulaireDefinition } from "./definition/form-definition";
import { StringMap } from "../stringmap";
import { FormulaireNode } from "./formulaire-node";
import { FieldsetContainer } from "./fieldset-container";

export class FieldSet extends FormulaireElement implements Observer {
    /**
     * Nub associé
     */
    private _nub: Nub;

    /**
     * dictionnaire de traduction
     */
    private _localisation: StringMap;

    /**
     * fichier de configuration
     */
    private _jsonConfig: {};

    constructor(parent: FormulaireNode) {
        super(parent);
    }

    public get nub(): Nub {
        return this._nub;
    }

    get label(): string {
        let label = this._label;
        if (this.parent instanceof FieldsetContainer) {
            label += " n°" + String(this.parent.getFieldsetPosition(this) + 1);
        }
        return label;
    }

    public setNub(sn: Nub, update: boolean = true) {
        this._nub = sn;
        this.setParentNubForAllFields();
        if (update) {
            this.updateFields();
        }
    }

    private addField(f: Field) {
        if (! f) {
            throw new Error("FieldSet.addField() : argument incorrect (undefined)");
        }
        this.kids.push(f);
    }

    public get hasInputs(): boolean {
        for (const f of this.kids) {
            if (f instanceof NgParameter) {
                return true;
            }
        }
        return false;
    }

    public getInput(i: number): NgParameter {
        let n = 0;
        for (const f of this.kids) {
            if (f instanceof NgParameter) {
                if (n === i) {
                    return f;
                }
                n++;
            }
        }
    }

    private parse_select(json: {}): SelectField {
        const res: SelectField = new SelectField(this);
        res.parseConfig(json);
        res.addObserver(this);
        return res;
    }

    public get properties(): Props {
        return this.nub.properties;
    }

    private get sectionProperties(): Props {
        const section = this.nub.getChildren()[0];
        if (section) {
            return section.properties;
        } else {
            return new Props();
        }
    }

    /**
     * @param inSection if true, will look for the required property in the Nub's section (children[0])
     */
    private getPropValue(key: string, inSection: boolean = false): any {
        if (inSection) {
            return this.sectionProperties.getPropValue(key);
        } else {
            return this.properties.getPropValue(key);
        }
    }

    public setPropValue(key: string, val: any): boolean {
        return this.properties.setPropValue(key, val, this);
    }

    private getNubParamFromSymbol(symbol: string): ParamDefinition {
        return this._nub.getParameter(symbol);
    }

    public get jsonConfig(): {} {
        return this._jsonConfig;
    }

    private setParentNubForAllFields() {
    }

    /**
     * crée un input
     * @param json definition de l'input, extrait du fichier de conf du module de calcul
     * @param node_type_filter filtre sur le type de noeud (input créé si undefined ou égal)
     * @param default_radio_config config du radio fixé/à varier/à calculer
     */
    private parse_input(json: {}): NgParameter {
        const input_id: string = json["id"];
        let res: NgParameter;
        let nubParam: ParamDefinition;
        try {
            nubParam = this.getNubParamFromSymbol(input_id);
            if (nubParam.visible) {
                res = new NgParameter(nubParam, this);
                res.parseConfig(json);
            }
        } catch (e) {
            // parameter does not exist in Nub, it's normal, simply do not show it (silent fail)
        }
        return res;
    }

    private parseFields() {
        const fields = this._jsonConfig["fields"];
        for (const field_index in fields) {
            let field = fields[field_index];
            // detect short input syntax
            if (typeof field === "string") {
                field = {
                    "id": field
                };
            }
            const type = field["type"] || "input"; // "input" by default
            let param: Field;

            switch (type) {
                case "input":
                    param = this.parse_input(field);
                    // potentiellement undefined car certaines définitions de FieldSet comportent des paramètres
                    // qui ne sont pas tous affichés en même temps (cf. ouvrages //)
                    if (param) {
                        this.addField(param);
                    }
                    break;

                case "select":
                    param = this.parse_select(field);
                    this.addField(param);
                    break;

            }
        }
    }

    private clearFields() {
        for (const n of this.kids) {
            n.removeObserver(this);
        }

        this.clearKids();
    }

    public updateLocalisation(loc?: StringMap) {
        if (! loc) {
            loc = this._localisation;
        } else {
            this._localisation = loc;
        }

        if (loc) {
            super.updateLocalisation(loc);
        }
    }

    /**
     * Reflects all properties values in the interface, through the values of the <select> fields
     */
    public updateFields() {
        this.clearFields();
        this.parseFields();
        this.updateLocalisation();

        // MAJ des selects avec les valeurs actuelles des propriétés
        // spécifique à chaque modul de calcul, à revoir
        switch (this._confId) {

            case "fs_ouvrage":
                this.setSelectValueFromProperty("select_structure", "structureType");
                this.setSelectValueFromProperty("select_loidebit", "loiDebit");
                break;

            case "fs_section":
                // property nodeType is stored in the section, not in the parent Nub
                this.setSelectValueFromProperty("select_section", "nodeType", true);
                break;

            case "fs_param_calc":
                this.setSelectValueFromProperty("select_resolution", "methodeResolution");
                break;

            case "fs_target_data": // courbe de remous
                this.setSelectValueFromProperty("select_target", "varCalc");
                break;

            case "fs_pass_type": // macro-rugo complexe
                this.setSelectValueFromProperty("select_pass_type", "inclinedApron");
                break;
        }
    }

    /**
     * Reflects a property value in the interface, through the value of a <select> field, if this select exists
     * @param inSection if true, will look for the required property in the Nub's section (children[0])
     */
    private setSelectValueFromProperty(selectId: string, propertyKey: string, inSection: boolean = false) {
        const selectField: SelectField = this.getFormulaireNodeById(selectId) as SelectField;
        if (selectField) {
            let propVal: any = this.getPropValue(propertyKey, inSection);
            if (propVal === undefined) {
                propVal = ""; // clodo bullet-proof loading
            }
            const selectElement = selectField.getSelectedEntryFromValue(propVal);
            try {
                selectField.setValue(selectElement);
            } catch (e) {
                console.error(`setSelectValueFromProperty: cannot set value ${propVal} on <select> ${selectId}`);
            }
        }
    }

    /**
     * Sets Nub default property from config file, unless this property is already set
     */
    private setPropertyValueFromConfig(json: {}, configKey: string, propertyKey: string, enumClass?) {
        const configValue: string = json[configKey];
        const currentValue = this.properties.getPropValue(propertyKey);
        if (configValue && (currentValue === undefined)) {
            let formalValue =  configValue;
            if (enumClass) {
                formalValue = enumClass[configValue];
            }
            this.setPropValue(propertyKey, formalValue);
        }
    }

    /**
     * Sets Nub default properties from config file, unless properties values are already set
     * (when deserialising an existing Nub)
     * @TODO default values should be held by model, not interface !
     */
    public parseConfig(json: {}, data?: {}) {
        this._jsonConfig = json;

        this._confId = json["id"];
        this._helpLink = json["help"];

        const parentForm = this.parentForm as FormulaireDefinition;
        const ct: string = json["calcType"];
        const currentCt = this.properties.getPropValue("calcType");
        const calc_type: CalculatorType = currentCt ? currentCt : (ct ? CalculatorType[ct] : parentForm.calculatorType);
        this.setPropValue("calcType", calc_type);

        const dnt: string = json["defaultNodeType"];
        const currentNt = this.properties.getPropValue("nodeType");
        const node_type: ComputeNodeType = currentNt ? currentNt : (dnt ? ComputeNodeType[dnt] : parentForm.nodeType);
        this.setPropValue("nodeType", node_type);

        this.setPropertyValueFromConfig(json, "defaultStructType", "structureType");
        this.setPropertyValueFromConfig(json, "defaultLoiDebit", "loiDebit", LoiDebit);
        this.setPropertyValueFromConfig(json, "methodeResolution", "methodeResolution", MethodeResolution);
        this.setPropertyValueFromConfig(json, "varCalc", "varCalc");

        this.updateFields();
    }

    public getNodeParameter(symbol: string): NgParameter {
        for (const p of this.kids) {
            if (p instanceof NgParameter) {
                if (p.isDisplayed && p.symbol === symbol) {
                    return p;
                }
            }
        }
    }

    public getNodeParameterValue(symbol: string): number {
        const p = this.getNodeParameter(symbol);
        if (! p) {
            throw new Error(`FieldSet.getNodeParameterValue() : pas de paramètre ${symbol} trouvé`);
        }

        switch (p.radioState) {
            case ParamRadioConfig.FIX:
                return p.getValue();

            case ParamRadioConfig.VAR:
            case ParamRadioConfig.CAL:
                return undefined;
        }
    }

    /**
     * retourne la valeur actuellement sélectionnée d'un SelectField (qui doit être affiché)
     * @param selectFieldId id du SelectField
     * @returns valeur courante du select sans le préfixe
     */
    public getSelectedValue(selectFieldId: string): string {
        for (const p of this.kids) {
            if (p instanceof SelectField && p.isDisplayed && p.id === selectFieldId) {
                const value: string = p.getValue().value;
                return FormulaireElement.removePrefix(value, selectFieldId + "_");
            }
        }
    }

    // interface Observer

    update(sender: any, data: any) {
        if (data.action) {
            switch (data.action) {
                case "select":
                    const senderId: string = sender.id.replace(/\d+$/, "");
                    switch (senderId) {
                        case "select_section": // sections paramétrées
                            this.setPropValue("nodeType", data.value.value);
                            break;
                        case "select_structure": // ouvrages parallèles
                            this.setPropValue("structureType", data.value.value);
                            break;
                        case "select_loidebit": // ouvrages parallèles et dérivés
                            this.setPropValue("loiDebit", data.value.value);
                            break;
                        case "select_resolution": // courbes de remous, méthode de résolution
                            this.setPropValue("methodeResolution", data.value.value);
                            break;
                        case "select_target": // courbes de remous, variable à calculer
                            this.setPropValue("varCalc", data.value.value);
                            break;
                        case "select_pass_type": // macro-rugo complexe
                            this.setPropValue("inclinedApron", data.value.value);
                            break;
                    }
                    break;
            }
        }
    }
}
